/* axpipe.c
 *
 * Copyright (C) 2001 RedHat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
/*
 * Modified for SIKIGAMI 2001,2002 Axe,Inc.
 */


#include "axpipe.h"
#include "axstream.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>

// The assorted AxPipe::AxPipe() constructors are described in sgpipe.h
void
AxPipe_init_valued( AxPipe * pipe,
				 const char *program, const char **argv,
	       int writable, int readable, int catch_stderr,
	       int stdin_fd, int stdout_fd, int stderr_fd)
{
    AxPipe_setup(pipe,program, argv, writable, readable, catch_stderr,
	  									stdin_fd, stdout_fd, stderr_fd);
}

void
AxPipe_setup( AxPipe * pipe,
				const char *program, const char **argv,
	      int writable, int readable,  int catch_stderr,
	      int stdin_fd, int stdout_fd, int stderr_fd)
{
    int socks[2];

    pipe->pid = 0;
    pipe->estatus = -1;

    if (!program || !argv)
    {
			pipe->errnum = EINVAL;
			return;
    }
    
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, socks))
    {
			pipe->errnum = errno;
			return;
    }

    // Fork.  The child process will run the program we're piping.
    pipe->pid = fork();
    if (pipe->pid < 0) {
			pipe->pid = 0;
			pipe->errnum = errno;
			close(socks[0]);
			close(socks[1]);
			return;
    }
    
    if (!pipe->pid) {
			close(socks[0]);
	
			if (writable)
	    	dup2(socks[1], 0); // writable means redirect child stdin
			else if (stdin_fd == -1)
	    	close(0);
			else
	    	dup2(stdin_fd, 0);

			if (readable)
	    	dup2(socks[1], 1); // readable means we redirect child stdout
			else if (stdout_fd == -1)
	    	close(1);
			else
	    	dup2(stdout_fd, 1);

			if (catch_stderr)
	    	dup2(socks[1], 2); // but catch_stderr does what you think
			else if (stderr_fd == -1)
	    	close(2);
			else
	    	dup2(stderr_fd, 2);
	
			fcntl(0, F_SETFD, 0);  // never close stdin
			fcntl(1, F_SETFD, 0);  // never close stdout
			fcntl(2, F_SETFD, 0);  // never close stderr
	
			if (!readable && stdout_fd != 1) {
	    	setsid();
	    	ioctl(1, TIOCSCTTY, 1);
			}
	
			execvp(program, (char **)argv);
			_exit(242);
    }

    // set non-blocking
    fcntl(socks[0], F_SETFL, O_RDWR|O_NONBLOCK);

    pipe->fd = socks[0];
    close(socks[1]);
}


// send the child process a signal
void
AxPipe_kill(AxPipe*pipe,int signum)
{
    if (pipe->pid)
			kill(pipe->pid, signum);
}

int
AxPipe_finish( AxPipe*pipe )
{
    while (!AxPipe_child_exited(pipe))
			usleep(100*1000);
    
    return AxPipe_exit_status(pipe);
}

int
AxPipe_child_exited( AxPipe * pipe )
{
    int status;
    pid_t dead_pid;
    
    if (!pipe->pid) return 1;
    
    dead_pid = waitpid(pipe->pid, &status, WNOHANG);
    if ( dead_pid != pipe->pid)
			return 0;
    pipe->estatus = status;
    pipe->pid = 0;
    return 1;
}


int
AxPipe_child_killed( AxPipe * pipe )
{
    int st = pipe->estatus;
    assert (WIFEXITED(st) || WIFSIGNALED(st));
    return WIFSIGNALED(st);
}

int
AxPipe_exit_status(AxPipe*pipe)
{
    int st = pipe->estatus;
    assert (WIFEXITED(st) || WIFSIGNALED(st));
    if (AxPipe_child_killed(pipe))
			return WTERMSIG(st);
    else
			return WEXITSTATUS(st);
}


// make sure our subtask ends up dead!
void
AxPipe_clear(AxPipe*pipe)
{
    int status, count;
    pid_t dead_pid;
    
    AxStream_close( (AxStream*)pipe );
    
    if (!pipe->pid) return;

    dead_pid = waitpid(pipe->pid, &status, WNOHANG);
    if (dead_pid == 0) {
	AxPipe_kill(pipe,SIGTERM);
	
	for (count = 20; count > 0; count--) {
  	  dead_pid = waitpid(pipe->pid, &status, WNOHANG);
    	  if (dead_pid == pipe->pid)
	    break;
	  usleep(100 * 1000);
	}
	
	if (dead_pid == 0) {
    	  AxPipe_kill(pipe,SIGKILL);
    	  waitpid(pipe->pid, &status, 0);
	}
    }
}
